#![no_std]
#![no_main]
#![allow(async_fn_in_trait)]
#![feature(impl_trait_in_assoc_type)]
#![feature(slice_split_once)]

use core::net::Ipv4Addr;
use cyw43_pio::{DEFAULT_CLOCK_DIVIDER, PioSpi};
use dhcparse::dhcpv4::{DhcpOption, Encode as _, MessageType, OpCode};
use dhcparse::{dhcpv4, v4_options};
use embassy_executor::Spawner;
use embassy_net::udp::{PacketMetadata, UdpMetadata, UdpSocket};
use embassy_net::{Config, DhcpConfig, IpEndpoint, StackResources};
use embassy_rp::bind_interrupts;
use embassy_rp::clocks::RoscRng;
use embassy_rp::gpio::{Level, Output};
use embassy_rp::peripherals::USB;
use embassy_rp::peripherals::{DMA_CH0, PIO0};
use embassy_rp::pio::{InterruptHandler as PioInterruptHandler, Pio};
use embassy_rp::usb::{Driver, InterruptHandler as UsbInterruptHandler};
use embassy_time::Timer;
use heapless::Vec;
use log::info;
use rand_core::RngCore;
use static_cell::StaticCell;
use {defmt_rtt as _, panic_probe as _};
use pico_website::unwrap;

mod socket;
mod game;

bind_interrupts!(struct Irqs {
    USBCTRL_IRQ => UsbInterruptHandler<USB>;
    PIO0_IRQ_0 => PioInterruptHandler<PIO0>;
});

#[embassy_executor::task]
async fn logger_task(driver: Driver<'static, USB>) {
    embassy_usb_logger::run!(1024, log::LevelFilter::Info, driver);
}

#[embassy_executor::task]
async fn cyw43_task(
    runner: cyw43::Runner<'static, Output<'static>, PioSpi<'static, PIO0, 0, DMA_CH0>>,
) -> ! {
    runner.run().await
}

#[embassy_executor::task]
async fn net_task(mut runner: embassy_net::Runner<'static, cyw43::NetDriver<'static>>) -> ! {
    runner.run().await
}

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    let p = embassy_rp::init(Default::default());
    let driver = Driver::new(p.USB, Irqs);
    spawner.spawn(logger_task(driver)).unwrap();
    let mut rng = RoscRng;

    let fw = include_bytes!("../cyw43-firmware/43439A0.bin");
    let clm = include_bytes!("../cyw43-firmware/43439A0_clm.bin");
    let pwr = Output::new(p.PIN_23, Level::Low);
    let cs = Output::new(p.PIN_25, Level::High);
    let mut pio = Pio::new(p.PIO0, Irqs);
    let spi = PioSpi::new(
        &mut pio.common,
        pio.sm0,
        DEFAULT_CLOCK_DIVIDER,
        pio.irq0,
        cs,
        p.PIN_24,
        p.PIN_29,
        p.DMA_CH0,
    );

    static STATE: StaticCell<cyw43::State> = StaticCell::new();
    let state = STATE.init(cyw43::State::new());
    let (net_device, mut control, runner) = cyw43::new(state, pwr, spi, fw).await;
    unwrap(spawner.spawn(cyw43_task(runner))).await;

    control.init(clm).await;
    control
        .set_power_management(cyw43::PowerManagementMode::PowerSave)
        .await;

    #[cfg(not(feature = "wifi-connect"))]
    // Use a link-local address for communication without DHCP server
    let config = Config::ipv4_static(embassy_net::StaticConfigV4 {
        address: embassy_net::Ipv4Cidr::new(embassy_net::Ipv4Address::new(192, 254, 0, 2), 16),
        dns_servers: heapless::Vec::new(),
        gateway: None,
    });

    #[cfg(feature = "wifi-connect")]
    let (wifi_conf, config) = {
        let wifi_conf = unwrap(serde_json_core::from_slice::<WifiConnectConf>(include_bytes!("../wifi.json"))).await.0;
        // Use a link-local address for communication without DHCP server
        // let config = Config::dhcpv4(embassy_net::DhcpConfig::default());
        let config = match wifi_conf.ip {
            Some(ip) => Config::ipv4_static(embassy_net::StaticConfigV4 {
                address: embassy_net::Ipv4Cidr::new(ip, 24),
                dns_servers: heapless::Vec::new(),
                gateway: None,
            }),
            None => Config::dhcpv4(DhcpConfig::default())
        };
        (wifi_conf, config)
    };


    // Generate random seed
    let seed = rng.next_u64();

    // Init network stack
    static RESOURCES: StaticCell<StackResources<4>> = StaticCell::new();
    let (stack, runner) = embassy_net::new(
        net_device,
        config,
        RESOURCES.init(StackResources::new()),
        seed,
    );

    unwrap(spawner.spawn(net_task(runner))).await;

    #[cfg(not(feature = "wifi-connect"))]
    //control.start_ap_open("cyw43", 5).await;
    control.start_ap_wpa2("cyw43", "password", 5).await;

    #[cfg(feature = "wifi-connect")]
    {
        loop {
            match control
                .join(wifi_conf.name, cyw43::JoinOptions::new(wifi_conf.password.as_bytes()))
                .await
            {
                Ok(_) => break,
                Err(err) => {
                    info!("join failed with status={}", err.status);
                }
            }
        }
        info!("Network joined!");
        info!("waiting for link...");
        stack.wait_link_up().await;
        // Wait for DHCP, not necessary when using static IP
        info!("waiting for DHCP...");
        stack.wait_config_up().await;
        // while !stack.is_config_up() {
        //     Timer::after_millis(100).await;
        // }
        info!("DHCP is now up!");
        info!(
            "ip : {}",
            unwrap(stack.config_v4().ok_or("no dhcp config"))
                .await
                .address
        )
    }

    let mut rx_buffer = [0; 4096];
    let mut tx_buffer = [0; 4096];
    let mut rx_meta = [PacketMetadata::EMPTY; 16];
    let mut tx_meta = [PacketMetadata::EMPTY; 16];
    let mut buf = [0; 4096];
    let mut res_buf = Vec::<u8, 4096>::new();
    loop {
        let mut socket = UdpSocket::new(stack, &mut rx_meta, &mut rx_buffer, &mut tx_meta, &mut tx_buffer);
        unwrap(socket.bind(67)).await;

        loop {
            // loop {
            //     if socket.may_recv() {
            //         break
            //     }
            // }
            // info!("may recv");
            // Timer::after_secs(0).await;

            let (n, ep) = unwrap(socket.recv_from(&mut buf).await).await;

            // if let Ok(s) = core::str::from_utf8(&buf[..n]) {
            //     info!("rxd str from {}: {}", ep, s);
            // } else {
            //     info!("rxd not str from {}: {:?}", ep, &buf[..n]);
            // }
            // Timer::after_secs(0).await;

            let msg = unwrap(dhcpv4::Message::new(&buf[..n])).await;

            let opts = unwrap(v4_options!(msg; MessageType required)).await;

            match opts {
                dhcpv4::MessageType::DISCOVER => {
                    res_buf.clear();
                    res_buf.push(2).unwrap(); // op
                    res_buf.push(1).unwrap(); // htype
                    res_buf.push(6).unwrap(); // hlen
                    res_buf.push(0).unwrap(); // hops
                    res_buf.extend_from_slice(&buf[4..8]).unwrap(); // xid
                    res_buf.extend_from_slice(&[0; 2]).unwrap(); // secs
                    res_buf.extend_from_slice(&buf[10..12]).unwrap(); // flags
                    res_buf.extend_from_slice(&[0; 4]).unwrap(); // ciaddr
                    res_buf.extend_from_slice(&[192, 254, 0, 12]).unwrap(); // yiaddr
                    res_buf.extend_from_slice(&[192, 254, 0, 2]).unwrap(); // siaddr
                    res_buf.extend_from_slice(&buf[24..28]).unwrap(); // giaddr
                    res_buf.extend_from_slice(&buf[28..44]).unwrap(); // chaddr
                    res_buf.extend_from_slice(&[0; 192]).unwrap(); // sname/file
                    res_buf.extend_from_slice(&[99, 130, 83, 99]).unwrap(); // magic cookie
                    // options
                    res_buf.extend_from_slice(&[53, 1, 2]).unwrap(); // message type
                    res_buf.extend_from_slice(&[1, 4, 255, 255, 255, 0]).unwrap(); // subnet mask
                    res_buf.extend_from_slice(&[3, 4, 192, 254, 0, 2]).unwrap(); // router
                    res_buf.extend_from_slice(&[6, 4, 0, 0, 0, 0]).unwrap(); // dns
                    res_buf.extend_from_slice(&[15, 1, 0]).unwrap(); // domain name
                    res_buf.extend_from_slice(&[26, 2, 0x5, 0xEA]).unwrap(); // mtu
                    res_buf.extend_from_slice(&[28, 4, 192, 254, 0, 255]).unwrap(); // domain name
                    res_buf.extend_from_slice(&[51, 4, 0, 0, 0x2, 0xBC]).unwrap(); // address lease time = 700s
                    res_buf.extend_from_slice(&[58, 4, 0, 0, 0x2, 0x58]).unwrap(); // renewal time = 600s
                    res_buf.extend_from_slice(&[59, 4, 0, 0, 0x2, 0x58]).unwrap(); // rebinding time = 600s
                    res_buf.extend_from_slice(&[43, 1, 0]).unwrap(); // vendor specific
                    let captive_portal_uri = "http://192.254.0.2:80/";
                    res_buf.extend_from_slice(&[114, captive_portal_uri.len() as u8]).unwrap(); // captive portal
                    res_buf.extend_from_slice(captive_portal_uri.as_bytes()).unwrap();
                    res_buf.extend_from_slice(&[108, 4, 0, 0, 0, 0]).unwrap(); // ipv6 only

                    for o in unwrap(unwrap(dhcpv4::Message::new(&res_buf)).await.options()).await {
                        info!("{:?}", match o {
                            Err(_) => break,
                            Ok(o) => o
                        });
                    }

                    unwrap(socket.send_to(&res_buf, UdpMetadata {
                        endpoint: IpEndpoint::new(Ipv4Addr::new(193, 254, 0, 12).into(), 68),
                        local_address: Some(Ipv4Addr::new(192, 254, 0, 2).into()),
                        meta: Default::default()
                    }).await).await;
                },
                _ => {}
            }
            for o in unwrap(msg.options()).await {
                info!("{:?}", match o {
                    Err(_) => break,
                    Ok(o) => o
                });
            }
            info!("{:?}", opts);
            Timer::after_secs(0).await;
            // unwrap(socket.send_to(&buf[..n], ep).await).await;
        }
    }

    // unwrap(spawner.spawn(socket::listen_task(stack, game::Team::Zero, 80))).await;
    // unwrap(spawner.spawn(socket::listen_task(stack, game::Team::One, 81))).await;
}

#[cfg(feature="wifi-connect")]
#[derive(serde::Deserialize)]
struct WifiConnectConf<'a> {
    name: &'a str,
    password: &'a str,
    ip: Option<Ipv4Addr>
}